# -*- coding: utf-8 -*-
"""
/***************************************************************************
 IncaPluginDockWidget
                                 A QGIS plugin
 KIP-INCA Natural Capital Accounting plugin.
 Generated by Plugin Builder: http://g-sherman.github.io/Qgis-Plugin-Builder/
                             -------------------
        begin                : 2021-05-12
        git sha              : $Format:%H$
        copyright            : (C) 2021 by VITO
        email                : thomas.danckaert@vito.be
 ***************************************************************************/
"""
import operator
import os
from functools import reduce

import geopandas as gpd
import yaml
from qgis.PyQt import QtGui, QtWidgets, QtCore, uic
from qgis.core import Qgis, QgsTask, QgsMessageLog, QgsApplication, QgsProject
from qgis.gui import QgsFileWidget
from qgis.utils import iface

import inca
import inca.carbon
import inca.cropprovision
import inca.nuts
import inca.recreation
import inca.soilretention
import inca.woodprovision
from inca.errors import Error, ConfigError
from inca.carbon.carbon_retention import STOCK, STOCK_DIFFERENCE, GAIN_LOSS
from inca.carbon.carbon_sequestration import EURO_TON_CO2, EURO_TON_C, TON_C, TON_CO2, KTON_C, KTON_CO2

from .qt_tools import expand_template, writeWidget, readWidget
from .wood_metadata_dialog import WoodMetadataDialog

FORM_CLASS, BASE_CLASS = uic.loadUiType(os.path.join(
    os.path.dirname(__file__), 'inca_plugin_dockwidget_base.ui'))


_inca_tasks = []


class IncaPluginDockWidget(BASE_CLASS, FORM_CLASS):

    closingPlugin = QtCore.pyqtSignal()

    def __init__(self, parent=None):
        super().__init__(parent)

        self.setupUi(self)

        self.metadata_editors = {
            'Wood Provision': WoodMetadataDialog
        }
        self.metadata = {}

        self.toolbar = QtWidgets.QToolBar()
        self.toolbar.setIconSize(iface.iconSize(dockedToolbar=True))
        self.loadact = QtWidgets.QAction(QtGui.QIcon(":/plugins/inca_plugin/mActionFileOpen.svg"), 'Load Configuration', self)
        self.loadact.triggered.connect(self.loadConfig)
        self.saveact = QtWidgets.QAction(QtGui.QIcon(":/plugins/inca_plugin/mActionFileSaveAs.svg"), 'Save Configuration', self)
        self.saveact.triggered.connect(self.saveConfig)
        self.toolbar.addAction(self.loadact)
        self.toolbar.addAction(self.saveact)
        self.toolbar.addSeparator()

        self.metadataact = QtWidgets.QAction('Edit metadata', self)
        self.metadataact.triggered.connect(self.edit_metadata)
        self.toolbar.addAction(self.metadataact)
        self.toolbar.addSeparator()

        self.runact = QtWidgets.QAction('Run', self)
        self.runact.triggered.connect(self.run)
        self.toolbar.addAction(self.runact)
        self.continue_run = QtWidgets.QCheckBox('Continue existing run')
        self.continue_run.setObjectName('continue_run')
        self.toolbar.addWidget(self.continue_run)

        self.toolbarLayout.addWidget(self.toolbar)

        self.set_up_config_templates()

        self.current_nuts_file = None
        self.NUTS_file_clicked = False
        self.eu_nuts = inca.nuts.get_eu_2020()
        for nuts_id in self.eu_nuts.index:
            self.selected_regions.addItem(nuts_id)
        self.selected_regions.selectAllOptions()

        nuts_line = self.NUTS.lineEdit()
        nuts_line.setNullValue('2021 EU members [NUTS 2021 level 0]')
        nuts_line.setValue(None)
        nuts_line.setShowClearButton(True)

        reclassification_line = self.land_classification.lineEdit()
        reclassification_line.setNullValue('INCA default mapping for CORINE data')
        reclassification_line.setValue(None)
        reclassification_line.setShowClearButton(True)

        typeofareacontribution_line = self.recreation_typeofareacontribution.lineEdit()
        typeofareacontribution_line.setNullValue('INCA default area type contribution factors')
        typeofareacontribution_line.setValue(None)

        areacontribution_line = self.recreation_areacontribution.lineEdit()
        areacontribution_line.setNullValue('Ignore')
        areacontribution_line.setValue(None)

        ecosystemweights_line = self.recreation_ecosystemweights.lineEdit()
        ecosystemweights_line.setNullValue('INCA default ecosystem weights')
        ecosystemweights_line.setValue(None)

        accessibilitymap_line = self.recreation_accessibilitymap.lineEdit()
        accessibilitymap_line.setNullValue('Ignore')
        accessibilitymap_line.setValue(None)

        accessibilityweights_line = self.recreation_accessibilityweights.lineEdit()
        accessibilityweights_line.setNullValue('Use map values as weights.')
        accessibilityweights_line.setValue(None)

        facilitiesmap_line = self.recreation_facilitiesmap.lineEdit()
        facilitiesmap_line.setNullValue('Ignore')
        facilitiesmap_line.setValue(None)

        facilitiesweights_line = self.recreation_facilitiesweights.lineEdit()
        facilitiesweights_line.setNullValue('Use map values as weights.')
        facilitiesweights_line.setValue(None)

        landscapeattractivenessmap_line = self.recreation_landscapeattractivenessmap.lineEdit()
        landscapeattractivenessmap_line.setNullValue('Ignore')
        landscapeattractivenessmap_line.setValue(None)

        landscapeattractivenessweights_line = self.recreation_landscapeattractivenessweights.lineEdit()
        landscapeattractivenessweights_line.setNullValue('Use map values as weights.')
        landscapeattractivenessweights_line.setValue(None)

        self.carbon_retention_operation_mode.addItem('Base-stock', STOCK)
        self.carbon_retention_operation_mode.addItem('Stock-difference', STOCK_DIFFERENCE)
        self.carbon_retention_operation_mode.addItem('Gain-loss', GAIN_LOSS)

        self.carbon_unit_GGE.addItem('ton of Carbon', TON_C)
        self.carbon_unit_GGE.addItem('kiloton of Carbon', KTON_C)
        self.carbon_unit_GGE.addItem('ton of CO₂', TON_CO2)
        self.carbon_unit_GGE.addItem('kiloton of CO₂', KTON_CO2)

        self.carbon_unit_sequ_price.addItem('Euro per ton of Carbon', EURO_TON_C)
        self.carbon_unit_sequ_price.addItem('Euro per ton of CO₂', EURO_TON_CO2)

        self.loadSettings()

        QgsProject.instance().readProject.connect(self.loadSettings)
        QgsProject.instance().writeProject.connect(self.saveSettings)

        self.service.currentTextChanged.connect(self.service_changed)

        self.connect_nuts_combobox()

    def connect_nuts_combobox(self):
        """Signals and slots to update the region selection combobox when a new NUTS shapefile is selected."""
        # We want to update the list of regions whenever a new NUTS shape file was set by the user, but only when the
        # user has finished editing (i.e. don't attempt to load a new NUTS shape while the user is still typing the
        # filename).  The file can be changed in two ways:
        # - user edits the text using the widget's LineEdit -> use the editingFinished signal for this
        # - user clicks the button and selects a file using the menu -> there's no simple method to detect this, so
        #   we use a slightly hacky workaround, see below:
        self.NUTS.lineEdit().editingFinished.connect(self.nuts_file_editingfinished)
        # Workaround to detect file updates using the button:
        # 1. find the QgsFileWidget button using findChildren, and listen for the 'pressed' signal to set a flag
        #    when the user has clicked the button.
        for x in self.NUTS.findChildren(QtWidgets.QAbstractButton):
            x.pressed.connect(self.nuts_file_buttonclicked)
        # 2. listen for the LineEdit's textChanged signal to detect updates to the filename, but only update the menu
        #    when the button clicked flag is set. (The LineEdit also sends a textChanged signal whenever the user types
        #    in the lineEdit)
        self.NUTS.lineEdit().textChanged.connect(self.nuts_file_textchanged)
        # 3. handle special case when the user clicks the button but then clicks cancel, so no textChanged signal is
        #    sent, and then edits the filename by typing.
        QgsApplication.instance().focusChanged.connect(self.focuschanged)

    def nuts_file_editingfinished(self):
        """Update NUTS region list after the user has edited the file name using the text field."""
        self.updateRegions()

    def nuts_file_textchanged(self):
        """Update NUTS region list after the user has selected a file using the QgsFileWidget button."""
        if self.NUTS_file_clicked:
            self.NUTS_file_clicked = False
            self.updateRegions()

    def nuts_file_buttonclicked(self):
        """Detect when the user opens the NUTS file selection dialog."""
        self.NUTS_file_clicked = True

    def focuschanged(self, old, new):
        # We handle the following special case:
        # 1) user clicks the file button for the NUTS shape
        # 2) user clicks cancel (thus, the file remains the same, but our NUTS_file_clicked flag will be set to True)
        # 3) user clicks the line edit and changes the text.
        #
        # -> if our NUTS_file_clicked flag is still True, editing the text will trigger an update of the NUTS region
        # list after the first keypress.  Therefore, we set the NUTS_file_clicked to False when the NUTS file lineEdit
        # receives focus.
        if new == self.NUTS.lineEdit():
            self.NUTS_file_clicked = False

    def service_changed(self, service_name):
        # If we have metadata editor for this service, enable the button
        self.metadataact.setEnabled(service_name in self.metadata_editors)


    def updateRegions(self):
        if self.current_nuts_file == self.NUTS.lineEdit().value():
            # Do nothing if the NUTS file after editing is the same as the one before.
            return

        if self.NUTS.lineEdit().isNull():
            # Use default: EU 2020 members
            self.selected_regions.clear()
            for nuts_id in self.eu_nuts.index:
                self.selected_regions.addItem(nuts_id)
        else:
            try:
                regions = gpd.read_file(self.NUTS.filePath())['NUTS_ID']
            except BaseException:
                QgsMessageLog.logMessage(f'Failed to read NUTS_ID\'s from shapefile "{self.NUTS.filePath()}"',
                                         level=Qgis.Critical)
                self.NUTS.lineEdit().setStyleSheet('border: 2px solid red')
                return
            self.selected_regions.clear()
            for nuts_id in sorted(regions):
                self.selected_regions.addItem(nuts_id)
        self.current_nuts_file = self.NUTS.lineEdit().value()
        self.selected_regions.selectAllOptions()
        self.selected_regions.repaint()

    def list_widgets(self, element):
        result = []
        if isinstance(element, QtCore.QObject):
            result.append(element)
        elif isinstance(element, dict):
            for x in element.values():
                result = result + self.list_widgets(x)
        elif isinstance(element, list):
            for x in element:
                result = result + self.list_widgets(x)
        else:
            pass
        return result

    def list_config_widgets(self):
        return self.list_widgets(self.main_config_template) + self.list_widgets(self.service_config_templates)

    def loadSettings(self):
        proj = QgsProject.instance()
        for widget in self.list_config_widgets():
            value, ok = proj.readEntry('inca', widget.objectName())
            if ok:
                writeWidget(widget, value)
            else:
                QgsMessageLog.logMessage(f'Project has no settings for widget {widget.objectName()}.', level=Qgis.Info)
        self.updateRegions()

    def saveSettings(self):
        proj = QgsProject.instance()
        for widget in self.list_config_widgets():
            proj.writeEntry('inca', widget.objectName(), str(readWidget(widget)))

    def closeEvent(self, event):
        self.closingPlugin.emit()
        event.accept()

    def saveConfig(self):
        """Save current ui state as a yaml config file."""
        filename, _ = QtWidgets.QFileDialog.getSaveFileName(self, "Save Config File", "config.yaml", "*.yaml")
        if not filename:
            return
        with open(filename, 'wt') as f:
            f.write(yaml.dump(self.make_inca_config()))

    def loadConfig(self):
        """Load a yaml config file."""
        filename, _ = QtWidgets.QFileDialog.getOpenFileName(self, "Load Config File", "",
                                                            "Config files (*.yaml);; All files (*.*)")
        if not filename:
            return

        try:
            with open(filename, 'rt') as f:
                config = yaml.safe_load(f)

            # Handle NUTS file and region list separately:
            nuts_val = config.pop('NUTS')
            if nuts_val == '':
                nuts_val = None
            self.NUTS.lineEdit().setValue(nuts_val)
            self.updateRegions()
            self.selected_regions.deselectAllOptions()
            self.selected_regions.setCheckedItems(config.pop('selected_regions'))
            # Handle year selection:
            self.year.setValue(config['years'][0])

            # remaining config values can be read automatically using the config templates
            main_template = {key: value for key, value in self.main_config_template.items()
                             if key not in ('NUTS', 'selected_regions', 'years')}
            service_template = self.service_config_templates[config['service']]

            self.load_template(config, main_template)
            self.load_template(config, service_template)
        except KeyError as e:
            QtWidgets.QMessageBox.critical(self, 'Error loading config', f'Missing config key {e}.')
        except BaseException as e:
            QtWidgets.QMessageBox.critical(self, 'Error loading config', f'Could not load config {filename}: {e}.')

        if 'metadata' in config:
            self.metadata[config['service']] = config['metadata'][config['years'][0]]

    def load_template(self, config, template):
        """Update the widgets listed in the template with values from config.

        We use recursion to process te nested dictionary structure of the template and config.  This is more or less
        the inverse of expand_template().

        NOTE: Assumes widget self.year was already set to the correct value."""
        if type(template) == dict:
            for key, value in template.items():
                # special case for yearly datasets ...
                if key == self.year:
                    self.load_template(config[self.year.value()], value)
                else:
                    self.load_template(config[key], value)
        elif type(template) in (int, bool, str):
            # We're only interested in the widgets referenced from the template.
            pass
        elif isinstance(template, QtCore.QObject):
            writeWidget(template, config)
        else:
            raise RuntimeError(f'Failed to process template {template}.')

    def set_up_config_templates(self):
        """Create a mapping between INCA config keys and corresponding interface widgets.

        This mapping is used to transform the UI state into an INCA config.
        """
        self.main_config_template = {
            'service': self.service,
            'output_dir': self.output_dir,
            'years': [self.year],
            'continue': self.continue_run,
            'NUTS': self.NUTS,
            'selected_regions': self.selected_regions,
            'land_cover': {self.year: self.land_cover},
            'land_classification': self.land_classification,
        }
        self.service_config_templates = {
            'Soil Retention': {'run_name': self.soil_retention_run_name,
                               'parameters': {'avg_soil_formation': self.soil_formation_factor,
                                              'scaling_N_content': self.N_content_scaling,
                                              'scaling_P_content': self.P_content_scaling},
                               'paths': {'paths_input_processed': {
                                   'avg_nutrient_prices': self.average_nutrient_prices,
                                   'bulksoil_price': self.bulksoil_price,
                                   'cfactor': {self.year: self.cfactor},
                                   'retention_rate': self.soil_retention_rate,
                                   'k-factor': self.kfactor,
                                   'ls-factor': self.lsfactor,
                                   'p-factor': self.pfactor,
                                   'r-factor': self.rfactor,
                                   'topsoil_N_content': self.topsoil_N,
                                   'topsoil_P_content': self.topsoil_P}},
                               'run': {'region_short': 'total_extent'}},
            'Crop Provisioning': {
                'run_name': self.crop_provisioning_run_name,
                'ecocon': self.ecocon,
                'values': self.values,
                'production': self.production,
                'proxy': self.crop_provisioning_proxy},
            'Global Climate Regulation': {
                'run_name': self.carbon_run_name,
                'parameters': {
                    'sequ_operation_mode': self.carbon_sequestration_operation_mode,
                    'reten_operation_mode': self.carbon_retention_operation_mode,
                    'unit_GGE': self.carbon_unit_GGE,
                    'unit_sequ_price': self.carbon_unit_sequ_price,
                },
                'land_classification_L2': self.carbon_land_classification,
                'data': {
                    'env_air_gge': self.carbon_env_air_gge,
                    'productivity_map': {self.year: self.carbon_productivity_map},
                    'management_map': {self.year: self.carbon_management_map},
                    'unmanaged_forest_map': {self.year: self.carbon_unmanaged_map},
                    'unmanaged_forest_table': self.carbon_unmanaged_table,
                    'biomass_table': self.carbon_biomass_table,
                    'carbon_price_table': self.carbon_price_table,
                    'deflator': self.carbon_deflator}},
            'Recreation': {
                'run_name': self.recreation_run_name,
                'data': {
                    'admin_shapefile_nuts': self.recreation_nuts2_regions,
                    'base_visits_data': self.recreation_overnightstays,
                    'typeAreaContributionTable': self.recreation_typeofareacontribution,
                    'dataAreaContributionTable': self.recreation_areacontribution,
                    'step3_ecosystemTypeAttractiveness_table': self.recreation_ecosystemweights,
                    'step3_accessibility_table': self.recreation_accessibilityweights,
                    'step3_facilities_table': self.recreation_facilitiesweights,
                    'step3_attractiveness_table': self.recreation_landscapeattractivenessweights,
                    'step3_accessibility_map': {self.year: self.recreation_accessibilitymap},
                    'step3_facilities_map': {self.year: self.recreation_facilitiesmap},
                    'step3_attractiveness_map': {self.year: self.recreation_landscapeattractivenessmap},
                }
            },
            'Wood Provision': {
                'run_name': self.wood_run_name,
                'for_vol_efa': self.wood_for_vol_efa,
                'for_sup_cp': self.wood_for_sup_cp,
                'for_basic': self.wood_for_basic,
                'proxy': {self.year: self.wood_proxy},
                'proxy_weights': {self.year: self.wood_proxy_weights},
                'deflator': self.wood_deflator
            }}

    def make_template(self):
        """Create an INCA config template based on the current ui settings."""
        service = self.service.currentText()

        # Combine the main config template with the specific template for the selected service:
        return {
            **self.main_config_template,
            **self.service_config_templates[service]
        }

    def make_inca_config(self):
        """Create an INCA config dict based on the current ui settings."""
        # Read settings from widgets:
        config_dict = expand_template(self.make_template())
        # Add metadata for this service if it exists:
        config_dict['metadata'] = {self.year.value(): self.metadata.get(config_dict['service'], None)}
        return config_dict

    def expand_year(self, template):
        """Recursively replace all occurrences of the self.year widget as a dict key by the year value.

        This is needed so we can look up widgets by their config path when we catch a
        inca.checked_config.ConfigException."""
        if type(template) == dict:  # nested dictionary
            result = {}
            for key, value in template.items():
                result_val = self.expand_year(value)
                if key is self.year:
                    result[self.year.value()] = result_val
                else:
                    result[key] = result_val
            return result
        else:
            return template

    def run(self):
        if not self.edit_metadata():  # User clicked cancel -> don't start run.
            return

        config = self.make_inca_config()
        widget_dict = self.expand_year(self.make_template())

        runtypes = {'Wood Provision': inca.woodprovision.WoodRun,
                    'Crop Provisioning': inca.cropprovision.CropRun,
                    'Global Climate Regulation': inca.carbon.GlobalClimateRun,
                    'Recreation': inca.recreation.RecreationRun,
                    'Soil Retention': inca.soilretention.SoilRun}

        year = self.year.value()
        output_rasters = {'Wood Provision':
                              [os.path.join('maps', f'wood-provision_map_use_1000-m3_{year}.tif')],
                          'Recreation':
                              [os.path.join('maps', f'recreation_map_supply_amountOvernightStays-foreign_{year}.tif'),
                               os.path.join('maps', f'recreation_map_supply_amountOvernightStays-national_{year}.tif'),
                               os.path.join('maps', f'recreation_map_supply_amountOvernightStays_{year}.tif')],
                          'Soil Retention':
                              [os.path.join('results', f'soil-retention_map_use_tonnes_{year}.tif')],
                          'Global Climate Regulation':
                              [os.path.join('maps', f'carbon_map_use_tonnes_{year}.tif')]
                          }

        service = self.service.currentText()
        task = IncaTask(f'{service} run {config["run_name"]}', config, widget_dict, runtypes[service],
                        output_rasters=output_rasters.get(service, []))
        _inca_tasks.append(task)

        QgsMessageLog.logMessage('Submitting INCA task', level=Qgis.Info)
        QgsApplication.taskManager().addTask(task)

    def edit_metadata(self):
        service_name = self.service.currentText()
        editor = self.metadata_editors.get(service_name)
        if editor is None:  # No editor for this service -> nothing to be done
            return True
        dialog = editor(self)  # pass window instance to editor constructor so we have access to widgets from there
        if dialog.exec() == QtWidgets.QDialog.Accepted:
            self.metadata[service_name] = expand_template(dialog.metadata_template())
            return True
        else:  # 'Cancel' clicked
            return False

class IncaTask(QgsTask):

    def __init__(self, description, config, widget_dict, runtype, output_rasters=[]):
        super().__init__(description)
        self.config = config
        self.widget_dict = widget_dict
        self.runtype = runtype
        self.output_rasters = output_rasters

    def run(self):
        try:
            run = self.runtype(self.config)
            run.start(progress_callback=self.setProgress)
            return True
        except Exception as e:
            self.exception = e
            return False

    def finished(self, result):
        if result:
            QgsMessageLog.logMessage('INCA task completed', level=Qgis.Info)
            for raster in self.output_rasters:
                path = os.path.join(self.config['output_dir'], self.config['run_name'], raster)
                iface.addRasterLayer(path)
        else:
            if self.exception is None:
                QgsMessageLog.logMessage('INCA task failed for unknown reason')
            else:
                QgsMessageLog.logMessage(f'INCA task raised an exception: {self.exception}', level=Qgis.Critical)
                try:
                    raise self.exception
                except ConfigError as e:
                    widget = reduce(operator.getitem, e.path, self.widget_dict)
                    if isinstance(widget, QgsFileWidget):  # TODO clean up!
                        widget = widget.lineEdit()
                    widget.setStyleSheet('border: 1px solid red')
                    QtWidgets.QMessageBox.warning(iface.mainWindow(), 'INCA configuration error', e.message)
                    widget.setStyleSheet('')
                except Error as e:
                    QtWidgets.QMessageBox.warning(iface.mainWindow(), 'INCA error', str(e))
                except BaseException as e:
                    QtWidgets.QMessageBox.critical(iface.mainWindow(), 'INCA unexpected error',
                                                   f'Something went wrong: "{e}".  Please refer to the '
                                                   f'log file at {inca.get_logfile()} for more details.')
        _inca_tasks.remove(self)
